"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from datetime import datetime, timedelta, time
from decimal import Decimal, ROUND_HALF_DOWN
from typing import Optional, List, Dict

from pytz import UTC

from biweeklybudget.db import db_session, upsert_record
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.account import Account
from biweeklybudget.models.plaid_items import PlaidItem
from biweeklybudget.models.plaid_accounts import PlaidAccount
from biweeklybudget.utils import plaid_client, dtnow

from plaid.models import (
    ItemGetRequest, TransactionsGetRequest, TransactionsGetRequestOptions
)

logger = logging.getLogger(__name__)


class PlaidUpdateResult:
    """Describes the result of updating a single account via Plaid."""

    def __init__(
        self, item: PlaidItem, success: bool, updated: int, added: int,
        exc: Optional[Exception], stmt_ids: Optional[List[int]]
    ):
        """
        Store the result of an update.

        :param item: the PlaidItem that this update pertains to
        :param success: whether the update succeeded or not
        :param updated: count of updated transactions
        :param added: count of added transactions
        :param exc: exception encountered, if any
        :param stmt_ids: list of added Statement IDs
        """
        self.item = item
        self.success = success
        self.updated = updated
        self.added = added
        self.exc = exc
        self.stmt_ids = stmt_ids

    @property
    def as_dict(self):
        return {
            'item_id': self.item.item_id,
            'success': self.success,
            'exception': str(self.exc),
            'statement_ids': self.stmt_ids,
            'added': self.added,
            'updated': self.updated
        }


class PlaidUpdater:

    def __init__(self):
        self.client = plaid_client()

    @classmethod
    def available_items(cls):
        """
        Return a list of :py:class:`~.PlaidItem` objects that can be updated via
        Plaid.

        :return: PlaidItem that can be updated via Plaid
        :rtype: list of :py:class:`~.PlaidItem` objects
        """
        return db_session.query(PlaidItem).order_by(
            PlaidItem.institution_name
        ).all()

    def update(self, items=None, days=30):
        """
        Update account balances and transactions from Plaid, for either all
        Plaid Items that are available or a specified list of Item IDs.

        :param items: a list of :py:class:`~.PlaidItem` objects to update
        :type items: list or None
        :param days: number of days of transactions to get from Plaid
        :type days: int
        :return: list of :py:class:`~.PlaidUpdateResult` instances
        :rtype: list
        """
        if items is None:
            items = self.available_items()
        logger.debug(
            'Running Plaid update for %d items: %s',
            len(items), items
        )
        result = []
        for item in items:
            result.append(self._do_item(item, days=days))
        return result

    def _do_item(self, item, days):
        """
        Request transactions from Plaid for one Item. Update balances and
        transactions for each Account in that item.

        :param item: the item to update
        :type item: PlaidItem
        :param days: number of days of transactions to get from Plaid
        :type days: int
        :rtype: PlaidUpdateResult
        """
        logger.info('Plaid update for %s', item)
        try:
            end_date: datetime = dtnow()
            start_date: datetime = end_date - timedelta(days=days)
            iteminfo = self.client.item_get(
                ItemGetRequest(access_token=item.access_token)
            )
            logger.info(
                'Item %s transactions status: %s', item,
                iteminfo.get('status', {}).get('transactions')
            )
            logger.debug(
                'Downloading Plaid transactions for item: %s from %s to %s',
                item, start_date, end_date
            )
            txns: List[dict]
            accts: Dict
            txns, accts = self._get_transactions(
                item.access_token, start_date, end_date
            )
            accounts: Dict[str, PlaidAccount] = {}
            pa: PlaidAccount
            for pa in db_session.query(PlaidAccount).filter(
                PlaidAccount.item_id == item.item_id
            ).all():
                accounts[pa.account_id] = pa
            stmt_ids: List[int] = []
            added: int = 0
            updated: int = 0
            txns_per_acct: Dict[str, list] = {}
            for t in txns:
                if t['account_id'] not in txns_per_acct:
                    txns_per_acct[t['account_id']] = []
                txns_per_acct[t['account_id']].append(t)
            for plaid_account_id in accts.keys():
                plaid_acct: PlaidAccount = accounts[plaid_account_id]
                acct: Optional[Account] = plaid_acct.account
                if acct is None:
                    logger.warning(
                        'Plaid item %s returned transactions for %s, but it is '
                        'not mapped to an Account.', item, plaid_acct
                    )
                    continue
                sid, a, u = self._stmt_for_acct(
                    acct, accts[plaid_account_id],
                    txns_per_acct.get(plaid_account_id, []), end_date
                )
                added += a
                updated += u
                stmt_ids.append(sid)
            item.last_updated = dtnow()
            db_session.add(item)
            db_session.commit()
            return PlaidUpdateResult(
                item, True, updated, added, None, stmt_ids
            )
        except Exception as ex:
            logger.error(
                'Exception encountered when updating item: %s (%s)',
                item.institution_name, item.item_id, exc_info=True
            )
            return PlaidUpdateResult(
                item, False, 0, 0, ex, None
            )

    def _get_transactions(
            self, access_token: str, start_dt: datetime, end_dt: datetime
    ):
        kwargs = {
            'access_token': access_token,
            'start_date': start_dt.date(),
            'end_date': end_dt.date(),
            'options': TransactionsGetRequestOptions()
        }
        txns: List[dict] = []
        accts: dict = {}
        req = TransactionsGetRequest(**kwargs)
        logger.debug('Issuing transactions get request')
        resp = self.client.transactions_get(req)
        txns = resp['transactions']
        for acct in resp['accounts']:
            accts[acct['account_id']] = acct
        logger.debug(
            'Got %d transactions; expecting %d total',
            len(txns), resp['total_transactions']
        )
        while len(txns) < resp['total_transactions']:
            kwargs['options'] = TransactionsGetRequestOptions(
                offset=len(txns)
            )
            req = TransactionsGetRequest(**kwargs)
            logger.debug(
                'Issuing transactions get request with offset=%d', len(txns)
            )
            resp = self.client.transactions_get(req)
            txns.extend(resp['transactions'])
            for acct in resp['accounts']:
                accts[acct['account_id']] = acct
        return txns, accts

    def _stmt_for_acct(
        self, account: Account, plaid_acct_info: dict, plaid_txns: List[dict],
        end_dt: datetime
    ):
        """
        Put Plaid transaction data to the DB

        :param account: the account to update
        :param plaid_acct_info: dict of account information from Plaid
        :param txns: list of transactions from Plaid
        :param end_dt: current time, as of when transactions were retrieved
        """
        stmt = OFXStatement(
            account_id=account.id,
            filename=f'Plaid_{account.name}_{int(end_dt.timestamp())}.ofx',
            file_mtime=end_dt,
            as_of=end_dt,
            currency=plaid_acct_info['balances']['iso_currency_code'],
            acctid=plaid_acct_info['mask']
        )
        stmt.bankid = account.plaid_account.plaid_item.institution_id
        if account.plaid_account.account_type == 'credit':
            stmt.type = 'CreditCard'
            self._update_bank_or_credit(
                end_dt, account, plaid_acct_info, plaid_txns, stmt
            )
        elif account.plaid_account.account_type == 'depository':
            stmt.type = 'Bank'
            self._update_bank_or_credit(
                end_dt, account, plaid_acct_info, plaid_txns, stmt
            )
        elif account.plaid_account.account_type == 'investment':
            stmt.type = 'Investment'
            self._update_investment(end_dt, account, plaid_acct_info, stmt)
        elif account.plaid_account.account_type == 'loan':
            # For now, this should work...
            stmt.type = 'Investment'
            self._update_investment(end_dt, account, plaid_acct_info, stmt)
        else:
            raise RuntimeError(
                'ERROR: Unknown account type: ' +
                account.plaid_account.account_type
            )
        count_new, count_upd = self._new_updated_counts()
        db_session.commit()
        logger.info('Account "%s" - inserted %d new OFXTransaction(s), updated '
                    '%d existing OFXTransaction(s)',
                    account.name, count_new, count_upd)
        logger.debug('Done updating OFX in DB')
        return stmt.id, count_new, count_upd

    def _new_updated_counts(self):
        """
        Return integer counts of the number of :py:class:`~.OFXTransaction`
        objects that have been created and updated.

        :return: 2-tuple of new OFXTransactions created, OFXTransactions updated
        :rtype: tuple
        """
        count_new = 0
        count_upd = 0
        for obj in db_session.dirty:
            if isinstance(obj, OFXTransaction):
                count_upd += 1
        for obj in db_session.new:
            if isinstance(obj, OFXTransaction):
                count_new += 1
        return count_new, count_upd

    def _update_bank_or_credit(
        self, end_dt: datetime, account: Account, plaid_acct_info: dict,
        plaid_txns: List[dict], stmt: OFXStatement
    ):
        logger.debug('Generating statement for credit account')
        stmt.as_of = end_dt
        stmt.ledger_bal_as_of = end_dt
        stmt.ledger_bal = Decimal(
            plaid_acct_info['balances']['current']
        ).quantize(
            Decimal('.01'), rounding=ROUND_HALF_DOWN
        )
        if plaid_acct_info['balances'].get('available', None) is not None:
            stmt.avail_bal = Decimal(
                plaid_acct_info['balances']['available']
            ).quantize(Decimal('.01'), rounding=ROUND_HALF_DOWN)
            stmt.avail_bal_as_of = end_dt
        stmt.currency = plaid_acct_info['balances']['iso_currency_code']
        db_session.add(stmt)
        account.set_balance(
            overall_date=stmt.as_of,
            ledger=stmt.ledger_bal,
            ledger_date=stmt.ledger_bal_as_of,
            avail=stmt.avail_bal,
            avail_date=stmt.avail_bal_as_of
        )
        for pt in plaid_txns:
            if pt['pending']:
                logger.info('Skipping pending transaction: %s', pt)
                continue
            kwargs = {
                'amount': Decimal(str(pt['amount'])).quantize(
                    Decimal('.01'), rounding=ROUND_HALF_DOWN
                ),
                'date_posted': datetime.combine(
                    pt["date"], time(0, 0, 0), tzinfo=UTC
                ),
                'fitid': pt['payment_meta']['reference_number'],
                'name': pt['name'],
                'account_id': account.id,
                'statement': stmt
            }
            if account.negate_ofx_amounts:
                kwargs['amount'] = kwargs['amount'] * Decimal('-1.0')
            if kwargs['fitid'] is None:
                kwargs['fitid'] = pt['transaction_id']
            upsert_record(
                OFXTransaction,
                ['account_id', 'fitid'],
                **kwargs
            )

    def _update_investment(
        self, end_dt: datetime, account: Account, plaid_acct_info: dict,
        stmt: OFXStatement
    ):
        logger.debug('Generating statement for investment account')
        stmt.as_of = end_dt
        stmt.ledger_bal = Decimal(
            plaid_acct_info['balances']['current']
        ).quantize(Decimal('.01'), rounding=ROUND_HALF_DOWN)
        stmt.ledger_bal_as_of = end_dt
        stmt.currency = plaid_acct_info['balances']['iso_currency_code']
        db_session.add(stmt)
        account.set_balance(
            overall_date=stmt.as_of,
            ledger=stmt.ledger_bal,
            ledger_date=stmt.ledger_bal_as_of
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from datetime import timedelta
from decimal import Decimal
from dateutil.relativedelta import relativedelta
from calendar import monthrange

from biweeklybudget.models.account import Account, AcctType

logger = logging.getLogger(__name__)


class InterestHelper(object):

    def __init__(self, db_sess, increases={}, onetimes={}):
        """
        Initialize interest calculation helper.

        :param db_sess: Database Session
        :type db_sess: sqlalchemy.orm.session.Session
        :param increases: dict of :py:class:`datetime.date` to
          :py:class:`decimal.Decimal` for new max payment amount to take effect
          on the specified date.
        :type increases: dict
        :param onetimes: dict of :py:class:`datetime.date` to
          :py:class:`decimal.Decimal` for additional amounts to add to the first
          maximum payment on or after the given date
        :type onetimes: dict
        """
        self._sess = db_sess
        self._accounts = self._get_credit_accounts()
        self._statements = self._make_statements(self._accounts)
        self._increases = increases
        self._onetimes = onetimes

    @property
    def accounts(self):
        """
        Return a dict of `account_id` to :py:class:`~.Account` for all Credit
        type accounts with OFX data present.

        :return: dict of account_id to Account instance
        :rtype: dict
        """
        return self._accounts

    def _get_credit_accounts(self):
        """
        Return a dict of `account_id` to :py:class:`~.Account` for all Credit
        type accounts with OFX data present.

        :return: dict of account_id to Account instance
        :rtype: dict
        """
        accts = self._sess.query(Account).filter(
            Account.acct_type.__eq__(AcctType.Credit),
            Account.is_active.__eq__(True)
        ).all()
        res = {a.id: a for a in accts}
        return res

    def _make_statements(self, accounts):
        """
        Make :py:class:`~.CCStatement` instances for each account; return a
        dict of `account_id` to CCStatement instance.

        :param accounts: dict of (int) account_id to Account instance
        :type accounts: dict
        :return: dict of (int) account_id to CCStatement instance
        :rtype: dict
        """
        res = {}
        for a_id, acct in accounts.items():
            icls = INTEREST_CALCULATION_NAMES[acct.interest_class_name]['cls'](
                acct.effective_apr
            )
            bill_period = _BillingPeriod(acct.balance.ledger_date.date())
            min_pay_cls = MIN_PAYMENT_FORMULA_NAMES[
                acct.min_payment_class_name]['cls']()
            res[a_id] = CCStatement(
                icls,
                abs(acct.balance.ledger),
                min_pay_cls,
                bill_period,
                end_balance=abs(acct.balance.ledger),
                interest_amt=acct.last_interest_charge
            )
        logger.debug('Statements: %s', res)
        return res

    @property
    def min_payments(self):
        """
        Return a dict of `account_id` to minimum payment for the latest
        statement, for each account.

        :return: dict of `account_id` to minimum payment (Decimal)
        :rtype: dict
        """
        res = {}
        for a_id, stmt in self._statements.items():
            res[a_id] = stmt.minimum_payment
        logger.debug('Minimum payments by account_id: %s', res)
        return res

    def calculate_payoffs(self):
        """
        Calculate payoffs for each account/statement.

        :return: dict of payoff information. Keys are payoff method names.
          Values are dicts, with keys "description" (str description of the
          payoff method), "doc" (the docstring of the class), and "results".
          The "results" dict has integer `account_id` as the key, and values are
          dicts with keys "payoff_months" (int), "total_payments" (Decimal),
          "total_interest" (Decimal) and ``next_payment`` (Decimal).
        :rtype: dict
        """
        res = {}
        max_total = sum(list(self.min_payments.values()))
        for name in sorted(PAYOFF_METHOD_NAMES.keys()):
            cls = PAYOFF_METHOD_NAMES[name]['cls']
            klass = cls(
                max_total, increases=self._increases, onetimes=self._onetimes
            )
            if not cls.show_in_ui:
                continue
            res[name] = {
                'description': PAYOFF_METHOD_NAMES[name]['description'],
                'doc': PAYOFF_METHOD_NAMES[name]['doc']
            }
            try:
                res[name]['results'] = self._calc_payoff_method(klass)
            except Exception as ex:
                res[name]['error'] = str(ex)
                logger.error('Minimum payment method %s failed: %s',
                             name, ex)
        return res

    def _calc_payoff_method(self, cls):
        """
        Calculate payoffs using one method.

        :param cls: payoff method class
        :type cls: biweeklybudget.interest._PayoffMethod
        :return: Dict with integer `account_id` as the key, and values are
          dicts with keys "payoff_months" (int), "total_payments" (Decimal),
          "total_interest" (Decimal), "next_payment" (Decimal).
        :rtype: dict
        """
        balances = {
            x: self._statements[x].principal for x in self._statements.keys()
        }
        res = {}
        calc = calculate_payoffs(cls, list(self._statements.values()))
        for idx, result in enumerate(calc):
            a_id = list(self._statements.keys())[idx]
            res[a_id] = {
                'payoff_months': result[0],
                'total_payments': result[1],
                'total_interest': result[1] - balances[a_id],
                'next_payment': result[2]
            }
        return res


class _InterestCalculation(object):

    #: Human-readable string name of the interest calculation type.
    description = None

    def __init__(self, apr):
        """
        :param apr: Annual Percentage Rate as a decimal
        :type apr: decimal.Decimal
        """
        self._apr = apr

    def __repr__(self):
        return '<%s(decimal.Decimal(\'%s\'))>' % (
            self.__class__.__name__, self.apr
        )

    @property
    def apr(self):
        return self._apr

    def calculate(self, principal, first_d, last_d, transactions={}):
        """
        Calculate compound interest for the specified principal.

        :param principal: balance at beginning of statement period
        :type principal: decimal.Decimal
        :param first_d: date of beginning of statement period
        :type first_d: datetime.date
        :param last_d: last date of statement period
        :type last_d: datetime.date
        :param transactions: dict of datetime.date to float amount adjust
          the balance by on the specified dates.
        :type transactions: dict
        :return: dict describing the result: end_balance (float),
          interest_paid (float)
        :rtype: dict
        """
        raise NotImplementedError("Must implement in subclass")


class AdbCompoundedDaily(_InterestCalculation):
    """
    Average Daily Balance method, compounded daily (like American Express).
    """

    #: Human-readable string name of the interest calculation type.
    description = 'Average Daily Balance Compounded Daily (AmEx)'

    def calculate(self, principal, first_d, last_d, transactions={}):
        """
        Calculate compound interest for the specified principal.

        :param principal: balance at beginning of statement period
        :type principal: decimal.Decimal
        :param first_d: date of beginning of statement period
        :type first_d: datetime.date
        :param last_d: last date of statement period
        :type last_d: datetime.date
        :param transactions: dict of datetime.date to float amount adjust
          the balance by on the specified dates.
        :type transactions: dict
        :return: dict describing the result: end_balance (float),
          interest_paid (float)
        :rtype: dict
        """
        dpr = self._apr / Decimal(365.0)
        interest = Decimal(0.0)
        num_days = 0
        bal_total = Decimal(0.0)
        bal = principal

        d = first_d
        while d <= last_d:
            num_days += 1
            if d in transactions:
                bal += transactions[d]
            int_amt = bal * dpr
            interest += int_amt
            bal += int_amt
            bal_total += bal
            d += timedelta(days=1)
        adb = bal_total / Decimal(num_days)
        final = adb * self._apr * num_days / Decimal(365.0)
        bal += final * dpr
        return {
            'interest_paid': final,
            'end_balance': bal
        }


class SimpleInterest(_InterestCalculation):
    """
    Simple interest, charged on balance at the end of the billing period.
    """

    #: Human-readable string name of the interest calculation type.
    description = 'Interest charged once on the balance at end of period.'

    def calculate(self, principal, first_d, last_d, transactions={}):
        """
        Calculate compound interest for the specified principal.

        :param principal: balance at beginning of statement period
        :type principal: decimal.Decimal
        :param first_d: date of beginning of statement period
        :type first_d: datetime.date
        :param last_d: last date of statement period
        :type last_d: datetime.date
        :param transactions: dict of datetime.date to float amount adjust
          the balance by on the specified dates.
        :type transactions: dict
        :return: dict describing the result: end_balance (float),
          interest_paid (float)
        :rtype: dict
        """
        num_days = 0
        bal = principal

        d = first_d
        while d <= last_d:
            num_days += 1
            if d in transactions:
                bal += transactions[d]
            d += timedelta(days=1)
        final = bal * self._apr * num_days / Decimal(365.0)
        return {
            'interest_paid': final,
            'end_balance': bal + final
        }


class _BillingPeriod(object):

    #: human-readable string description of the billing period type
    description = None

    def __init__(self, end_date, start_date=None):
        """
        Construct a billing period that is defined by a number of days.

        :param end_date: end date of the billing period
        :type end_date: datetime.date
        :param start_date: start date for billing period; if specified, will
          override calculation of start date
        :type start_date: datetime.date
        """
        self._period_for_date = end_date
        if start_date is None:
            if end_date.day < 15:
                # if end date is < 15, period is month before end_date
                self._end_date = (end_date.replace(day=1) - timedelta(days=1))
                self._start_date = self._end_date.replace(day=1)
            else:
                # if end date >= 15, period is month containing end_date
                self._start_date = end_date.replace(day=1)
                self._end_date = end_date.replace(
                    day=(monthrange(
                        end_date.year, end_date.month
                    )[1])
                )
        else:
            self._start_date = start_date
            self._end_date = self._start_date.replace(
                day=(monthrange(
                    self._start_date.year, self._start_date.month
                )[1])
            )

    def __repr__(self):
        return '<BillingPeriod(%s, start_date=%s)>' % (
            self._end_date, self._start_date
        )

    @property
    def start_date(self):
        return self._start_date

    @property
    def end_date(self):
        return self._end_date

    @property
    def payment_date(self):
        period_length = (self._end_date - self._start_date).days
        return self._start_date + timedelta(days=int(period_length / 2))

    @property
    def next_period(self):
        """
        Return the next billing period after this one.

        :return: next billing period
        :rtype: _BillingPeriod
        """
        return _BillingPeriod(
            self._end_date + relativedelta(months=1),
            start_date=(self._end_date + timedelta(days=1))
        )

    @property
    def prev_period(self):
        """
        Return the previous billing period before this one.

        :return: previous billing period
        :rtype: _BillingPeriod
        """
        e = self._start_date - timedelta(days=1)
        return _BillingPeriod(e, start_date=e.replace(day=1))


class _MinPaymentFormula(object):

    #: human-readable string description of the formula
    description = None

    def __init__(self):
        pass

    def calculate(self, balance, interest):
        """
        Calculate the minimum payment for a statement with the given balance
        and interest amount.

        :param balance: balance amount for the statement
        :type balance: decimal.Decimal
        :param interest: interest charged for the statement period
        :type interest: decimal.Decimal
        :return: minimum payment for the statement
        :rtype: decimal.Decimal
        """
        raise NotImplementedError()


class MinPaymentAmEx(_MinPaymentFormula):
    """
    Interest on last statement plus 1% of balance,
    or $35 if balance is less than $35.
    """

    #: human-readable string description of the formula
    description = 'AmEx - Greatest of Interest Plus 1% of Principal, or $35'

    def __init__(self):
        super(MinPaymentAmEx, self).__init__()

    def calculate(self, balance, interest):
        """
        Calculate the minimum payment for a statement with the given balance
        and interest amount.

        :param balance: balance amount for the statement
        :type balance: decimal.Decimal
        :param interest: interest charged for the statement period
        :type interest: decimal.Decimal
        :return: minimum payment for the statement
        :rtype: decimal.Decimal
        """
        amt = interest + (balance * Decimal('.01'))
        if amt < 35:
            amt = Decimal('35')
        return amt


class MinPaymentDiscover(_MinPaymentFormula):
    """
    Greater of:
    - $35; or
    - 2% of the New Balance shown on your billing statement; or
    - $20, plus any of the following charges as shown on your billing statement:
    fees for any debt protection product that you enrolled in on or after
    2/1/2015; Interest Charges; and Late Fees.
    """

    #: human-readable string description of the formula
    description = 'Discover - Greatest of 2% of Principal, or $20 plus ' \
                  'Interest, or $35'

    def __init__(self):
        super(MinPaymentDiscover, self).__init__()

    def calculate(self, balance, interest):
        """
        Calculate the minimum payment for a statement with the given balance
        and interest amount.

        :param balance: balance amount for the statement
        :type balance: decimal.Decimal
        :param interest: interest charged for the statement period
        :type interest: decimal.Decimal
        :return: minimum payment for the statement
        :rtype: decimal.Decimal
        """
        options = [
            Decimal(35),
            balance * Decimal('0.02'),
            Decimal(20) + interest
        ]
        return max(options)


class MinPaymentCiti(_MinPaymentFormula):
    """
    Greater of:
    - $25;
    - The new balance, if it's less than $25;
    - 1 percent of the new balance, plus the current statement's interest
    charges or minimum interest charges, plus late fees;
    - 1.5% of the new balance, rounded to the nearest dollar amount.

    In all cases, add past fees and finance charges due, plus any amount in
    excess of credit line.
    """

    #: human-readable string description of the formula
    description = 'Citi - Greatest of 1.5% of Principal, or 1% of Principal ' \
                  'plus interest and fees, or $25, or Principal'

    def __init__(self):
        super(MinPaymentCiti, self).__init__()

    def calculate(self, balance, interest):
        """
        Calculate the minimum payment for a statement with the given balance
        and interest amount.

        :param balance: balance amount for the statement
        :type balance: decimal.Decimal
        :param interest: interest charged for the statement period
        :type interest: decimal.Decimal
        :return: minimum payment for the statement
        :rtype: decimal.Decimal
        """
        options = [
            25,
            (balance * Decimal('0.01')) + interest,
            round(balance * Decimal('0.015'))
        ]
        if balance < Decimal('25'):
            options.append(balance)
        return max(options)


class _PayoffMethod(object):
    """
    A payoff method for multiple cards; a method of figuring out how much to
    pay on each card, each month.
    """

    #: human-readable string name of the payoff method
    description = None

    def __init__(self, max_total_payment=None, increases={}, onetimes={}):
        """
        Initialize a payment method.

        :param max_total_payment: maximum total payment for all statements
        :type max_total_payment: decimal.Decimal
        :param increases: dict of :py:class:`datetime.date` to
          :py:class:`decimal.Decimal` for new max payment amount to take effect
          on the specified date.
        :type increases: dict
        :param onetimes: dict of :py:class:`datetime.date` to
          :py:class:`decimal.Decimal` for additional amounts to add to the first
          maximum payment on or after the given date
        :type onetimes: dict
        """
        self._max_total = max_total_payment
        self._increases = increases
        self._onetimes = onetimes

    def __repr__(self):
        return '<%s(%s, increases=%s, onetimes=%s)>' % (
            self.__class__.__name__, self._max_total, self._increases,
            self._onetimes
        )

    def max_total_for_period(self, period):
        """
        Given a :py:class:`~._BillingPeriod`, calculate the maximum total
        payment for that period, including both `self._max_total` and the
        increases and onetimes specified on the class constructor.

        :param period: billing period to get maximum total payment for
        :type period: _BillingPeriod
        :return: maximum total payment for the period
        :rtype: decimal.Decimal
        """
        res = self._max_total
        for inc_d in sorted(self._increases.keys(), reverse=True):
            if inc_d > period.payment_date:
                continue
            inc_amt = self._increases[inc_d]
            logger.debug('Found increase of %s starting on %s, applied to '
                         'period %s', inc_amt, inc_d, period)
            res = inc_amt
            break
        for ot_d, ot_amt in self._onetimes.items():
            if period.prev_period.payment_date < ot_d <= period.payment_date:
                logger.debug('Found onetime of %s on %s in period %s',
                             ot_amt, ot_d, period)
                res += ot_amt
        logger.debug('Period %s _max_total=%s max_total_for_period=%s',
                     period, self._max_total, res)
        return res

    def find_payments(self, statements):
        """
        Given a list of statements, return a list of payment amounts to make
        on each of the statements.

        :param statements: statements to pay, list of :py:class:`~.CCStatement`
        :type statements: list
        :return: list of payment amounts to make, same order as ``statements``
        :rtype: list
        """
        raise NotImplementedError()


class MinPaymentMethod(_PayoffMethod):
    """
    Pay only the minimum on each statement.
    """

    description = 'Minimum Payment Only'
    show_in_ui = True

    def find_payments(self, statements):
        """
        Given a list of statements, return a list of payment amounts to make
        on each of the statements.

        :param statements: statements to pay, list of :py:class:`~.CCStatement`
        :type statements: list
        :return: list of payment amounts to make, same order as ``statements``
        :rtype: list
        """
        return [s.minimum_payment for s in statements]


class FixedPaymentMethod(_PayoffMethod):
    """
    TESTING ONLY - pay the same amount on every statement.
    """

    description = 'TESTING ONLY - Fixed Payment for All Statements'
    show_in_ui = False

    def find_payments(self, statements):
        """
        Given a list of statements, return a list of payment amounts to make
        on each of the statements.

        :param statements: statements to pay, list of :py:class:`~.CCStatement`
        :type statements: list
        :return: list of payment amounts to make, same order as ``statements``
        :rtype: list
        """
        return [self._max_total for _ in statements]


class HighestBalanceFirstMethod(_PayoffMethod):
    """
    Pay statements off from highest to lowest balance.
    """

    description = 'Highest to Lowest Balance'
    show_in_ui = True

    def find_payments(self, statements):
        """
        Given a list of statements, return a list of payment amounts to make
        on each of the statements.

        :param statements: statements to pay, list of :py:class:`~.CCStatement`
        :type statements: list
        :return: list of payment amounts to make, same order as ``statements``
        :rtype: list
        """
        max_total = self.max_total_for_period(statements[0].billing_period)
        min_sum = sum([s.minimum_payment for s in statements])
        if min_sum > max_total:
            raise TypeError(
                'ERROR: Max total payment of %s is less than sum of minimum '
                'payments (%s)' % (max_total, min_sum)
            )
        max_bal = Decimal('0.00')
        max_idx = None
        for idx, stmt in enumerate(statements):
            if stmt.principal > max_bal:
                max_bal = stmt.principal
                max_idx = idx
        res = [None for _ in statements]
        max_pay = max_total - (
            min_sum - statements[max_idx].minimum_payment
        )
        for idx, stmt in enumerate(statements):
            if idx == max_idx:
                res[idx] = max_pay
            else:
                res[idx] = statements[idx].minimum_payment
        return res


class HighestInterestRateFirstMethod(_PayoffMethod):
    """
    Pay statements off from highest to lowest interest rate.
    """

    description = 'Highest to Lowest Interest Rate'
    show_in_ui = True

    def find_payments(self, statements):
        """
        Given a list of statements, return a list of payment amounts to make
        on each of the statements.

        :param statements: statements to pay, list of :py:class:`~.CCStatement`
        :type statements: list
        :return: list of payment amounts to make, same order as ``statements``
        :rtype: list
        """
        max_total = self.max_total_for_period(statements[0].billing_period)
        min_sum = sum([s.minimum_payment for s in statements])
        if min_sum > max_total:
            raise TypeError(
                'ERROR: Max total payment of %s is less than sum of minimum '
                'payments (%s)' % (max_total, min_sum)
            )
        max_apr = Decimal('0.00')
        max_idx = None
        for idx, stmt in enumerate(statements):
            if stmt.apr > max_apr:
                max_apr = stmt.apr
                max_idx = idx
        res = [None for _ in statements]
        max_pay = max_total - (
            min_sum - statements[max_idx].minimum_payment
        )
        for idx, stmt in enumerate(statements):
            if idx == max_idx:
                res[idx] = max_pay
            else:
                res[idx] = statements[idx].minimum_payment
        return res


class LowestBalanceFirstMethod(_PayoffMethod):
    """
    Pay statements off from lowest to highest balance, a.k.a. the "snowball"
    method.
    """

    description = 'Lowest to Highest Balance (a.k.a. Snowball Method)'
    show_in_ui = True

    def find_payments(self, statements):
        """
        Given a list of statements, return a list of payment amounts to make
        on each of the statements.

        :param statements: statements to pay, list of :py:class:`~.CCStatement`
        :type statements: list
        :return: list of payment amounts to make, same order as ``statements``
        :rtype: list
        """
        max_total = self.max_total_for_period(statements[0].billing_period)
        min_sum = sum([s.minimum_payment for s in statements])
        if min_sum > max_total:
            raise TypeError(
                'ERROR: Max total payment of %s is less than sum of minimum '
                'payments (%s)' % (max_total, min_sum)
            )
        min_bal = Decimal('+Infinity')
        min_idx = None
        for idx, stmt in enumerate(statements):
            if stmt.principal < min_bal:
                min_bal = stmt.principal
                min_idx = idx
        res = [None for _ in statements]
        min_pay = max_total - (
            min_sum - statements[min_idx].minimum_payment
        )
        for idx, stmt in enumerate(statements):
            if idx == min_idx:
                res[idx] = min_pay
            else:
                res[idx] = statements[idx].minimum_payment
        return res


class LowestInterestRateFirstMethod(_PayoffMethod):
    """
    Pay statements off from lowest to highest interest rate.
    """

    description = 'Lowest to Highest Interest Rate'
    show_in_ui = True

    def find_payments(self, statements):
        """
        Given a list of statements, return a list of payment amounts to make
        on each of the statements.

        :param statements: statements to pay, list of :py:class:`~.CCStatement`
        :type statements: list
        :return: list of payment amounts to make, same order as ``statements``
        :rtype: list
        """
        max_total = self.max_total_for_period(statements[0].billing_period)
        min_sum = sum([s.minimum_payment for s in statements])
        if min_sum > max_total:
            raise TypeError(
                'ERROR: Max total payment of %s is less than sum of minimum '
                'payments (%s)' % (max_total, min_sum)
            )
        min_apr = Decimal('+Infinity')
        min_idx = None
        for idx, stmt in enumerate(statements):
            if stmt.apr < min_apr:
                min_apr = stmt.apr
                min_idx = idx
        res = [None for _ in statements]
        min_pay = max_total - (
            min_sum - statements[min_idx].minimum_payment
        )
        for idx, stmt in enumerate(statements):
            if idx == min_idx:
                res[idx] = min_pay
            else:
                res[idx] = statements[idx].minimum_payment
        return res


def calculate_payoffs(payment_method, statements):
    """
    Calculate the amount of time (in years) and total amount of money required
    to pay off the cards associated with the given list of statements. Return a
    list of (`float` number of years, `decimal.Decimal` amount paid,
    `decimal.Decimal` first payment amount) tuples for each item in
    `statements`.

    :param payment_method: method used for calculating payment amount to make
      on each statement; subclass of _PayoffMethod
    :type payment_method: _PayoffMethod
    :param statements: list of :py:class:`~.CCStatement` objects to pay off.
    :type statements: list
    :return: list of (`float` number of billing periods, `decimal.Decimal`
      amount paid, `decimal.Decimal` first payment amount) tuples for each item
      in `statements`
    :rtype: list
    """
    def unpaid(s): return [x for x in s.keys() if s[x]['done'] is False]
    payoffs = {}
    logger.debug(
        'calculating payoff via %s for: %s', payment_method, statements
    )
    for idx, stmt in enumerate(statements):
        payoffs[stmt] = {
            'months': 0, 'amt': Decimal('0.0'), 'idx': idx, 'done': False,
            'next_pymt_amt': None
        }
    while len(unpaid(payoffs)) > 0:
        u = unpaid(payoffs)
        to_pay = payment_method.find_payments(u)
        for stmt, p_amt in dict(zip(u, to_pay)).items():
            if stmt.principal <= Decimal('0'):
                payoffs[stmt]['done'] = True
                continue
            if stmt.principal <= p_amt:
                payoffs[stmt]['done'] = True
                payoffs[stmt]['months'] += 1  # increment months
                payoffs[stmt]['amt'] += stmt.principal
                if payoffs[stmt]['next_pymt_amt'] is None:
                    payoffs[stmt]['next_pymt_amt'] = stmt.principal
                continue
            payoffs[stmt]['months'] += 1  # increment months
            payoffs[stmt]['amt'] += p_amt
            if payoffs[stmt]['next_pymt_amt'] is None:
                payoffs[stmt]['next_pymt_amt'] = p_amt
            new_s = stmt.pay(Decimal('-1') * p_amt)
            payoffs[new_s] = payoffs[stmt]
            del payoffs[stmt]
    res = []
    for s in sorted(payoffs, key=lambda x: payoffs[x]['idx']):
        tmp = (
            payoffs[s]['months'],
            payoffs[s]['amt'],
            payoffs[s]['next_pymt_amt']
        )
        if payoffs[s]['next_pymt_amt'] is None:
            tmp = (
                payoffs[s]['months'],
                payoffs[s]['amt'],
                Decimal('0.0')
            )
        res.append(tmp)
    return res


class CCStatement(object):
    """
    Represent a credit card statement (one billing period).
    """

    def __init__(self, interest_cls, principal, min_payment_cls, billing_period,
                 transactions={}, end_balance=None, interest_amt=None):
        """
        Initialize a CCStatement. At least one of `start_date` and `end_date`
        must be specified.

        :param interest_cls: Interest calculation method
        :type interest_cls: _InterestCalculation
        :param principal: starting principal for this billing period
        :type principal: decimal.Decimal
        :param min_payment_cls: Minimum payment calculation method
        :type min_payment_cls: _MinPaymentFormula
        :param billing_period: Billing period
        :type billing_period: _BillingPeriod
        :param transactions: transactions applied during this statement. Dict
          of :py:class:`datetime.date` to :py:class:`decimal.Decimal`.
        :type transactions: dict
        :param end_balance: the ending balance of the statement, if known. If
          not specified, this value will be calculated.
        :type end_balance: decimal.Decimal
        :param interest_amt: The amount of interest charged this statement. If
          not specified, this value will be calculated.
        :type interest_amt: decimal.Decimal
        """
        if not isinstance(billing_period, _BillingPeriod):
            raise TypeError(
                'billing_period must be an instance of _BillingPeriod'
            )
        self._billing_period = billing_period
        if not isinstance(interest_cls, _InterestCalculation):
            raise TypeError(
                'interest_cls must be an instance of _InterestCalculation'
            )
        self._interest_cls = interest_cls
        if not isinstance(min_payment_cls, _MinPaymentFormula):
            raise TypeError(
                'min_payment_cls must be an instance of _MinPaymentFormula'
            )
        self._min_pay_cls = min_payment_cls
        self._orig_principal = principal
        self._min_pay = None
        self._transactions = transactions
        self._principal = end_balance
        self._interest_amt = interest_amt
        if end_balance is None or interest_amt is None:
            res = self._interest_cls.calculate(
                principal, self._billing_period.start_date,
                self._billing_period.end_date, self._transactions
            )
            if end_balance is None:
                self._principal = res['end_balance']
            if interest_amt is None:
                self._interest_amt = res['interest_paid']

    def __repr__(self):
        return '<CCStatement(interest_cls=%s principal=%s min_payment_cls=%s ' \
               'transactions=%s end_balance=%s ' \
               'interest_amt=%s start_date=%s end_date=%s)>' % (
                   self._interest_cls, self._principal, self._min_pay_cls,
                   self._transactions, self._principal,
                   self._interest_amt, self.start_date, self.end_date
               )

    @property
    def principal(self):
        return self._principal

    @property
    def billing_period(self):
        """
        Return the Billing Period for this statement.

        :return: billing period for this statement
        :rtype: _BillingPeriod
        """
        return self._billing_period

    @property
    def interest(self):
        return self._interest_amt

    @property
    def start_date(self):
        return self._billing_period.start_date

    @property
    def end_date(self):
        return self._billing_period.end_date

    @property
    def apr(self):
        return self._interest_cls.apr

    @property
    def minimum_payment(self):
        """
        Return the minimum payment for the next billing cycle.

        :return: minimum payment for the next billing cycle
        :rtype: decimal.Decimal
        """
        return self._min_pay_cls.calculate(
            self._principal, self._interest_amt
        )

    def next_with_transactions(self, transactions={}):
        """
        Return a new CCStatement reflecting the next billing period, with a
        payment of `amount` applied to it.

        :param transactions: dict of transactions, `datetime.date` to `Decimal`
        :type transactions: dict
        :return: next period statement, with transactions applied
        :rtype: CCStatement
        """
        return CCStatement(
            self._interest_cls,
            self._principal,
            self._min_pay_cls,
            self._billing_period.next_period,
            transactions=transactions
        )

    def pay(self, amount):
        """
        Return a new CCStatement reflecting the next billing period, with a
        payment of `amount` applied to it at the middle of the period.

        :param amount: amount to pay during the next statement period
        :type amount: decimal.Decimal
        :return: next period statement, with payment applied
        :rtype: CCStatement
        """
        return self.next_with_transactions({
            self._billing_period.next_period.payment_date: amount
        })


def subclass_dict(klass):
    d = {}
    for cls in klass.__subclasses__():
        d[cls.__name__] = {
            'description': cls.description,
            'doc': cls.__doc__.strip(),
            'cls': cls
        }
    return d


#: Dict mapping interest calculation class names to their description and
#: docstring.
INTEREST_CALCULATION_NAMES = subclass_dict(_InterestCalculation)

#: Dict mapping Minimum Payment Formula class names to their description and
#: docstring.
MIN_PAYMENT_FORMULA_NAMES = subclass_dict(_MinPaymentFormula)

#: Dict mapping Payoff Method class names to their description and docstring.
PAYOFF_METHOD_NAMES = subclass_dict(_PayoffMethod)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from __future__ import with_statement
import os
from alembic import context
from sqlalchemy import pool, create_engine
from logging.config import fileConfig
import logging
import warnings
from biweeklybudget.settings import DB_CONNSTRING
from biweeklybudget.models.base import Base

logger = logging.getLogger(__name__)

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# only configure logging if used standalone, not when imported in app:
if len(logging.getLogger().handlers) < 1:
    # Interpret the config file for Python logging.
    # This line sets up loggers basically.
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def env_get_connstring():
    """Helper for migration tests"""
    try:
        conf = config.get_section_option('bwbTest', 'connstring', default=None)
    except Exception:
        conf = None
    if conf:
        logger.debug('Overriding connstring to: %s', conf)
        return conf
    logger.debug('Using default connstring from settings: %s', DB_CONNSTRING)
    return DB_CONNSTRING


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = env_get_connstring()
    context.configure(
        url=url, target_metadata=target_metadata, literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    echo = False
    if os.environ.get('SQL_ECHO', '') == 'true':
        echo = True
    connectable = create_engine(
        env_get_connstring(), poolclass=pool.NullPool,
        pool_pre_ping=('SQL_POOL_PRE_PING' in os.environ), echo=echo
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata
        )
        with context.begin_transaction():
            with warnings.catch_warnings():
                # Ignore warning from MariaDB MDEV-17544 when creating tables;
                # SQLAlchemy 1.3.13 names the primary keys, but MariaDB ignores
                # the names; MariaDB >= 10.4.7 now throws a warning for this.
                warnings.filterwarnings(
                    'ignore',
                    message=r'^\(1280, "Name.*ignored for PRIMARY key\."\)$',
                    category=Warning
                )
                context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""add Account negate_ofx_amounts field

Revision ID: a8f1228af250
Revises: 7477ff15e8c4
Create Date: 2017-04-19 22:20:31.876889

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a8f1228af250'
down_revision = '7477ff15e8c4'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('accounts', sa.Column('negate_ofx_amounts', sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column('accounts', 'negate_ofx_amounts')



from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import with_statement
try:
    # python 3
    from configparser import ConfigParser
except ImportError:
    # python 2
    from ConfigParser import ConfigParser
try:
    import keyring
    keyring.get_password('is-backend', 'configured?')
    KEYRING_AVAILABLE = True
except RuntimeError:
    # no keyring backend found
    KEYRING_AVAILABLE = False
except ImportError:
    KEYRING_AVAILABLE = False
except KeyError:
    KEYRING_AVAILABLE = False
import os
import os.path
import sys

from biweeklybudget.vendored.ofxclient.account import Account

try:
    DEFAULT_CONFIG = os.path.expanduser(os.path.join('~', 'ofxclient.ini'))
except:
    DEFAULT_CONFIG = None


class SecurableConfigParser(ConfigParser):
    """:py:class:`ConfigParser` subclass that knows how to store
    options marked as secure into the OS specific
    keyring/keychain.

    To mark an option as secure, the caller must call
    'set_secure' at least one time for the particular
    option and from then on it will be seen as secure
    and will be stored / retrieved from the keychain.

    Example::

      from biweeklybudget.vendored.ofxclient.config import SecurableConfigParser

      # password will not be saved in the config file

      c = SecurableConfigParser()
      c.add_section('Info')
      c.set('Info','username','bill')
      c.set_secure('Info','password','s3cre7')
      with open('config.ini','wb') as fp:
        c.write(fp)
    """

    _secure_placeholder = '%{secured}'

    def __init__(self, keyring_name='ofxclient',
                 keyring_available=KEYRING_AVAILABLE, **kwargs):
        if sys.version_info >= (3,):
            # python 3
            ConfigParser.__init__(self, interpolation=None)
        else:
            # python 2
            ConfigParser.__init__(self)
        self.keyring_name = keyring_name
        self.keyring_available = keyring_available
        self._unsaved = {}
        self.keyring_name = keyring_name

    def is_secure_option(self, section, option):
        """Test an option to see if it is secured or not.

        :param section: section id
        :type section: string
        :param option: option name
        :type option: string
        :rtype: boolean
        otherwise.
        """
        if not self.has_section(section):
            return False
        if not self.has_option(section, option):
            return False
        if ConfigParser.get(self, section, option) == self._secure_placeholder:
            return True
        return False

    def has_secure_option(self, section, option):
        """See is_secure_option"""
        return self.is_secure_option(section, option)

    def items(self, section):
        """Get all items for a section. Subclassed, to ensure secure
        items come back with the unencrypted data.

        :param section: section id
        :type section: string
        """
        items = []
        for k, v in ConfigParser.items(self, section):
            if self.is_secure_option(section, k):
                v = self.get(section, k)
            if v == '!!False!!':
                v = False
            items.append((k, v))
        return items

    def secure_items(self, section):
        """Like items() but only return secure items.

        :param section: section id
        :type section: string
        """
        return [x
                for x in self.items(section)
                if self.is_secure_option(section, x[0])]

    def set(self, section, option, value):
        """Set an option value. Knows how to set options properly marked
        as secure."""
        if value is False:
            value = '!!False!!'
        if self.is_secure_option(section, option):
            self.set_secure(section, option, value)
        else:
            ConfigParser.set(self, section, option, value)

    def set_secure(self, section, option, value):
        """Set an option and mark it as secure.

        Any subsequent uses of 'set' or 'get' will also
        now know that this option is secure as well.
        """
        if self.keyring_available:
            s_option = "%s%s" % (section, option)
            self._unsaved[s_option] = ('set', value)
            value = self._secure_placeholder
        ConfigParser.set(self, section, option, value)

    def get(self, section, option, *args):
        """Get option value from section. If an option is secure,
        populates the plain text."""
        if self.is_secure_option(section, option) and self.keyring_available:
            s_option = "%s%s" % (section, option)
            if self._unsaved.get(s_option, [''])[0] == 'set':
                res = self._unsaved[s_option][1]
            else:
                res = keyring.get_password(self.keyring_name, s_option)
        else:
            res = ConfigParser.get(self, section, option, *args)
        if res == '!!False!!':
            return False
        return res

    def remove_option(self, section, option):
        """Removes the option from ConfigParser as well as
        the secure storage backend
        """
        if self.is_secure_option(section, option) and self.keyring_available:
            s_option = "%s%s" % (section, option)
            self._unsaved[s_option] = ('delete', None)
        ConfigParser.remove_option(self, section, option)

    def write(self, *args):
        """See ConfigParser.write().  Also writes secure items to keystore."""
        ConfigParser.write(self, *args)
        if self.keyring_available:
            for key, thing in self._unsaved.items():
                action = thing[0]
                value = thing[1]
                if action == 'set':
                    keyring.set_password(self.keyring_name, key, value)
                elif action == 'delete':
                    try:
                        keyring.delete_password(self.keyring_name, key)
                    except:
                        pass
        self._unsaved = {}


class OfxConfig(object):
    """Default config file handler for other tools to use.

    This can read and write from the default config which is
    $USERS_HOME/ofxclient.ini

    :param file_name: absolute path to a config file (optional)
    :type file_name: string or None

    Example usage::

      from biweeklybudget.vendored.ofxclient.config import OfxConfig
      from biweeklybudget.vendored.ofxclient import Account

      a = Account()

      c = OfxConfig(file_name='/tmp/new.ini')
      c.add_account(a)
      c.save()

      account_list = c.accounts()
      one_account  = c.account( a.local_id() )
    """

    def __init__(self, file_name=None):

        self.secured_field_names = [
            'institution.username',
            'institution.password'
        ]

        f = file_name or DEFAULT_CONFIG
        if f is None:
            raise ValueError('file_name is required')
        self._load(f)

    def reload(self):
        """Reload the config file from disk"""
        return self._load()

    def accounts(self):
        """List of confgured :py:class:`ofxclient.Account` objects"""
        return [self._section_to_account(s)
                for s in self.parser.sections()]

    def encrypted_accounts(self):
        return [a
                for a in self.accounts()
                if self.is_encrypted_account(a.local_id())]

    def unencrypted_accounts(self):
        return [a
                for a in self.accounts()
                if not self.is_encrypted_account(a.local_id())]

    def account(self, id):
        """Get :py:class:`ofxclient.Account` by section id"""
        if self.parser.has_section(id):
            return self._section_to_account(id)
        return None

    def add_account(self, account):
        """Add Account to config (does not save)"""
        serialized = account.serialize()
        section_items = flatten_dict(serialized)
        section_id = section_items['local_id']

        if not self.parser.has_section(section_id):
            self.parser.add_section(section_id)

        for key in sorted(section_items):
            self.parser.set(section_id, key, section_items[key])

        self.encrypt_account(id=section_id)

        return self

    def encrypt_account(self, id):
        """Make sure that certain fields are encrypted."""
        for key in self.secured_field_names:
            value = self.parser.get(id, key)
            self.parser.set_secure(id, key, value)
        return self

    def is_encrypted_account(self, id):
        """Are all fields for the account id encrypted?"""
        for key in self.secured_field_names:
            if not self.parser.is_secure_option(id, key):
                return False
        return True

    def remove_account(self, id):
        """Add Account from config (does not save)"""
        if self.parser.has_section(id):
            self.parser.remove_section(id)
            return True
        return False

    def save(self):
        """Save changes to config file"""
        with open(self.file_name, 'w') as fp:
            self.parser.write(fp)
        return self

    def _load(self, file_name=None):
        self.parser = None

        file_name = file_name or self.file_name

        if not os.path.exists(file_name):
            with open(file_name, 'a'):
                os.utime(file_name, None)

        self.file_name = file_name

        conf = SecurableConfigParser()
        with open(self.file_name) as f:
            if hasattr(conf, 'read_file'):
                # python 3
                conf.read_file(f)
            else:
                # python 2
                conf.readfp(f)
        self.parser = conf

        return self

    def _section_to_account(self, section):
        section_items = dict(self.parser.items(section))
        serialized = unflatten_dict(section_items)
        return Account.deserialize(serialized)


def unflatten_dict(dict, prefix=None, separator='.'):
    ret = {}
    for k, v in dict.items():
        key_parts = k.split(separator)

        if len(key_parts) == 1:
            ret[k] = v
        else:
            first = key_parts[0]
            rest = key_parts[1:]
            temp = ret.setdefault(first, {})
            for idx, part in enumerate(rest):
                if (idx+1) == len(rest):
                    temp[part] = v
                else:
                    temp = temp.setdefault(part, {})
    return ret


def flatten_dict(dict_, prefix=None, separator='.'):
    ret = {}
    for k, v in dict_.items():
        if prefix:
            flat_key = separator.join([prefix, k])
        else:
            flat_key = k
        if isinstance(v, dict):
            deflated = flatten_dict(v, prefix=flat_key)
            for dk, dv in deflated.items():
                ret[dk] = dv
        else:
            ret[flat_key] = v
    return ret



"""add Vehicle and FuelLog models

Revision ID: 8627a80874a6
Revises: ed827c515edb
Create Date: 2017-07-01 09:21:51.583666

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '8627a80874a6'
down_revision = 'ed827c515edb'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'vehicles',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=254), nullable=True),
        sa.Column('is_active', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_vehicles')),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'fuellog',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('date', sa.Date(), nullable=True),
        sa.Column('vehicle_id', sa.Integer(), nullable=True),
        sa.Column('odometer_miles', sa.Integer(), nullable=True),
        sa.Column('reported_miles', sa.SmallInteger(), nullable=True),
        sa.Column('calculated_miles', sa.SmallInteger(), nullable=True),
        sa.Column('level_before', sa.SmallInteger(), nullable=True),
        sa.Column('level_after', sa.SmallInteger(), nullable=True),
        sa.Column('fill_location', sa.String(length=254), nullable=True),
        sa.Column(
            'cost_per_gallon',
            sa.Numeric(precision=10, scale=4),
            nullable=True
        ),
        sa.Column(
            'total_cost',
            sa.Numeric(precision=10, scale=4),
            nullable=True
        ),
        sa.Column(
            'gallons',
            sa.Numeric(precision=10, scale=4),
            nullable=True
        ),
        sa.Column(
            'reported_mpg',
            sa.Numeric(precision=10, scale=4),
            nullable=True
        ),
        sa.Column(
            'calculated_mpg',
            sa.Numeric(precision=10, scale=4),
            nullable=True
        ),
        sa.Column('notes', sa.String(length=254), nullable=True),
        sa.ForeignKeyConstraint(
            ['vehicle_id'],
            ['vehicles.id'],
            name=op.f('fk_fuellog_vehicle_id_vehicles')
        ),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_fuellog')),
        mysql_engine='InnoDB'
    )


def downgrade():
    op.drop_table('fuellog')
    op.drop_table('vehicles')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from datetime import datetime
from pytz import UTC

from ofxparse import AccountType, OfxParser
from biweeklybudget.db import db_session, upsert_record
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.account import Account
from biweeklybudget.utils import dtnow
from biweeklybudget.ofxapi.exceptions import DuplicateFileException

logger = logging.getLogger(__name__)


class OfxApiLocal(object):

    def __init__(self, db_sess):
        """
        Initialize a new Local OFX API client, when running ofxgetter or
        ofxbackfiller with direct database access, or used to back the OFX
        HTTP API.

        :param db_sess: active database session to use for queries
        :type db_sess: sqlalchemy.orm.session.Session
        """
        self._db = db_sess

    def get_accounts(self):
        """
        Query the database for all
        :py:attr:`ofxgetter-enabled
        <biweeklybudget.models.account.Account.for_ofxgetter>`
        :py:class:`Accounts <biweeklybudget.models.account.Account>` that have
        a non-empty
        :py:attr:`biweeklybudget.models.account.Account.ofxgetter_config` and a
        non-None
        :py:attr:`biweeklybudget.models.account.Account.vault_creds_path`.
        Return a dict of string
        :py:attr:`Account name <biweeklybudget.models.account.Account.name>` to
        dict with keys:

        - ``vault_path`` - :py:attr:`~.Account.vault_creds_path`
        - ``config`` - :py:attr:`~.Account.ofxgetter_config`
        - ``id`` - :py:attr:`~.Account.id`
        - ``cat_memo`` - :py:attr:`~.Account.ofx_cat_memo_to_name`

        :return: dict of account names to configuration
        :rtype: dict
        """
        result = {}
        for acct in self._db.query(Account).filter(
            Account.for_ofxgetter
        ).order_by(Account.name).all():
            if acct.vault_creds_path is None and acct.ofxgetter_config == {}:
                continue
            result[acct.name] = {
                'vault_path': acct.vault_creds_path,
                'config': acct.ofxgetter_config,
                'id': acct.id,
                'cat_memo': acct.ofx_cat_memo_to_name
            }
        logger.debug('Query found %d ofxgetter-enabled Accounts', len(result))
        return result

    def update_statement_ofx(self, acct_id, ofx, mtime=None, filename=None):
        """
        Update a single statement for the specified account, from an OFX file.

        :param acct_id: Account ID that statement is for
        :type acct_id: int
        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param mtime: OFX file modification time (or current time)
        :type mtime: datetime.datetime
        :param filename: OFX file name
        :type filename: str
        :returns: 3-tuple of the int ID of the
          :py:class:`~biweeklybudget.models.ofx_statement.OFXStatement`
          created by this run, int count of new :py:class:`~.OFXTransaction`
          created, and int count of :py:class:`~.OFXTransaction` updated
        :rtype: tuple
        :raises: :py:exc:`RuntimeError` on error parsing OFX or unknown account
          type; :py:exc:`~.DuplicateFileException` if the file (according to the
          OFX signon date/time) has already been recorded.
        """
        logger.info(
            'Updating Account %d with OFX Statement filename="%s" (mtime %s)',
            acct_id, filename, mtime
        )
        acct = db_session.query(Account).get(acct_id)
        if mtime is None:
            mtime = dtnow()
        if hasattr(ofx, 'status') and ofx.status['severity'] == 'ERROR':
            raise RuntimeError("OFX Error: %s" % vars(ofx))
        stmt = self._create_statement(acct, ofx, mtime, filename)
        if ofx.account.type == AccountType.Bank:
            stmt.type = 'Bank'
            s = self._update_bank_or_credit(acct, ofx, stmt)
        elif ofx.account.type == AccountType.CreditCard:
            stmt.type = 'CreditCard'
            s = self._update_bank_or_credit(acct, ofx, stmt)
        elif ofx.account.type == AccountType.Investment:
            s = self._update_investment(acct, ofx, stmt)
        else:
            raise RuntimeError("Don't know how to update AccountType %d",
                               ofx.account.type)
        count_new, count_upd = self._new_updated_counts()
        db_session.commit()
        return s.id, count_new, count_upd

    def _new_updated_counts(self):
        """
        Return integer counts of the number of :py:class:`~.OFXTransaction`
        objects that have been created and updated.

        :return: 2-tuple of new OFXTransactions created, OFXTransactions updated
        :rtype: tuple
        """
        count_new = 0
        count_upd = 0
        for obj in db_session.dirty:
            if isinstance(obj, OFXTransaction):
                count_upd += 1
        for obj in db_session.new:
            if isinstance(obj, OFXTransaction):
                count_new += 1
        return count_new, count_upd

    def _create_statement(self, acct, ofx, mtime, filename):
        """
        Create an OFXStatement for this OFX file. If one already exists with
        the same account and filename, raise DuplicateFileException.

        :param acct: the Account this statement is for
        :type acct: biweeklybudget.models.account.Account
        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param mtime: OFX file modification time (or current time)
        :type mtime: datetime.datetime
        :param filename: OFX file name
        :type filename: str
        :return: the OFXStatement object
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        :raises: DuplicateFileException
        """
        ofx_date = OfxParser.parseOfxDateTime(
            ofx.signon.dtserver).replace(tzinfo=UTC)
        stmt = db_session.query(OFXStatement).filter(
            OFXStatement.account_id == acct.id,
            OFXStatement.filename == filename
        ).first()
        if stmt is not None:
            logger.debug('Found existing statement with same as_of date, '
                         'id=%d; raising DuplicateFileException()', stmt.id)
            raise DuplicateFileException(acct.id, filename, stmt.id)
        logger.debug('Creating new OFXStatement account_id=%s filename=%s '
                     'as_of=%s', acct.id, filename, ofx_date)
        a = OFXStatement(
            account_id=acct.id,
            filename=filename,
            file_mtime=mtime,
            as_of=ofx_date,
            currency=ofx.account.curdef,
            acctid=ofx.account.account_id
        )
        if ofx.account.institution is not None:
            a.bankid = ofx.account.institution.fid
        if ofx.account.routing_number is not None:
            a.routing_number = ofx.account.routing_number
        if ofx.account.account_type is not None:
            a.acct_type = ofx.account.account_type
        if hasattr(ofx.account, 'brokerid'):
            a.brokerid = ofx.account.brokerid
        return a

    def _update_bank_or_credit(self, acct, ofx, stmt):
        """
        Update a single OFX file for this Bank or Credit account.

        :param acct: the Account this statement is for
        :type acct: biweeklybudget.models.account.Account
        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param stmt: the OFXStatement for this statement
        :type stmt: biweeklybudget.models.ofx_statement.OFXStatement
        :returns: the OFXStatement object
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        """
        # Note that as of 0.16, OfxParser returns tz-naive UTC datetimes
        logger.debug('Updating Bank/Credit account')
        if hasattr(ofx.account.statement, 'available_balance'):
            stmt.avail_bal = ofx.account.statement.available_balance
        if hasattr(ofx.account.statement, 'available_balance_date'):
            stmt.avail_bal_as_of = \
                ofx.account.statement.available_balance_date.replace(tzinfo=UTC)
        stmt.ledger_bal = ofx.account.statement.balance
        stmt.ledger_bal_as_of = \
            ofx.account.statement.balance_date.replace(tzinfo=UTC)
        db_session.add(stmt)
        acct.set_balance(
            overall_date=stmt.as_of,
            ledger=stmt.ledger_bal,
            ledger_date=stmt.ledger_bal_as_of,
            avail=stmt.avail_bal,
            avail_date=stmt.avail_bal_as_of
        )
        for txn in ofx.account.statement.transactions:
            try:
                kwargs = OFXTransaction.params_from_ofxparser_transaction(
                    txn, acct.id, stmt, cat_memo=acct.ofx_cat_memo_to_name
                )
            except RuntimeError as ex:
                logger.error(ex)
                continue
            upsert_record(
                OFXTransaction,
                ['account_id', 'fitid'],
                **kwargs
            )
        return stmt

    def _update_investment(self, acct, ofx, stmt):
        """
        Update a single OFX file for this Investment account.

        :param acct: the Account this statement is for
        :type acct: biweeklybudget.models.account.Account
        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param stmt: the OFXStatement for this statement
        :type stmt: biweeklybudget.models.ofx_statement.OFXStatement
        :returns: the OFXStatement object
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        """
        logger.debug('Updating Investment account')
        stmt.type = 'Investment'
        value = 0
        earliest_dt = datetime.max
        for pos in ofx.account.statement.positions:
            if hasattr(pos, 'date') and isinstance(pos.date, datetime):
                if pos.date < earliest_dt:
                    earliest_dt = pos.date
            if hasattr(pos, 'market_value'):
                value += pos.market_value
            else:
                value += (pos.units * pos.unit_price)
        if earliest_dt != datetime.max:
            stmt.ledger_bal_as_of = earliest_dt.replace(tzinfo=UTC)
        if value != 0:
            stmt.ledger_bal = value
        db_session.add(stmt)
        acct.set_balance(
            overall_date=stmt.as_of,
            ledger=stmt.ledger_bal,
            ledger_date=stmt.ledger_bal_as_of
        )
        return stmt



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import logging
import json

from alembic import command
from sqlalchemydiff import compare
from sqlalchemydiff.util import prepare_schema_from_models

from alembicverify.util import (
    get_current_revision,
    get_head_revision,
    prepare_schema_from_migrations,
)
from biweeklybudget.models.base import Base

import biweeklybudget.tests.migrations.alembic_helpers as ah

logger = logging.getLogger(__name__)


@pytest.mark.migrations
def test_upgrade_and_downgrade(uri_left, alembic_config_left):
    """Test all migrations up and down.

    Tests that we can apply all migrations from a brand new empty
    database, and also that we can remove them all.
    """
    alembic_config_left.set_section_option('bwbTest', 'connstring', uri_left)
    logger.info('Set alembic config bwbTest.connstring to: %s', uri_left)
    ah.load_premigration_sql(uri_left)
    engine, script = prepare_schema_from_migrations(
        uri_left, alembic_config_left
    )

    head = get_head_revision(alembic_config_left, engine, script)
    current = get_current_revision(alembic_config_left, engine, script)

    assert head == current

    while current is not None:
        command.downgrade(alembic_config_left, '-1')
        current = get_current_revision(alembic_config_left, engine, script)


@pytest.mark.migrations
def test_model_and_migration_schemas_are_the_same(
        uri_left, uri_right, alembic_config_left):
    """Compares the database schema obtained with all migrations against the
    one we get out of the models.
    """
    alembic_config_left.set_section_option('bwbTest', 'connstring', uri_left)
    logger.info('Set alembic config bwbTest.connstring to: %s', uri_left)
    ah.load_premigration_sql(uri_left)
    prepare_schema_from_migrations(uri_left, alembic_config_left)
    prepare_schema_from_models(uri_right, Base)

    result = compare(
        uri_left, uri_right,
        ignores=[
            'alembic_version',
            # for some reason, these constraints don't diff correctly,
            # likely due to creation order
            'accounts.cons.CONSTRAINT_1',
            'accounts.cons.CONSTRAINT_2',
            'accounts.cons.CONSTRAINT_3',
            'accounts.cons.CONSTRAINT_4',
            'budgets.cons.CONSTRAINT_1',
            'budgets.cons.CONSTRAINT_2',
            'budgets.cons.CONSTRAINT_3',
            'budgets.cons.CONSTRAINT_4',
            'ofx_trans.cons.CONSTRAINT_1',
            'ofx_trans.cons.CONSTRAINT_2',
            'ofx_trans.cons.CONSTRAINT_3',
            'ofx_trans.cons.CONSTRAINT_4',
            'ofx_trans.cons.CONSTRAINT_5',
            'reconcile_rules.cons.CONSTRAINT_1',
            'scheduled_transactions.cons.CONSTRAINT_1',
        ]
    )

    assert result.is_match is True, \
        'Differences (left is migrations, right is models):\n' \
        '%s' % json.dumps(
            result.errors, sort_keys=True, indent=4, separators=(',', ': ')
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from sqlalchemy import func
from decimal import Decimal

from biweeklybudget.db import db_session
from biweeklybudget.utils import dtnow, fmt_currency
from biweeklybudget.models.account import Account
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod

logger = logging.getLogger(__name__)


class NotificationsController(object):

    @staticmethod
    def num_stale_accounts(sess=None):
        """
        Return the number of accounts with stale data.

        @TODO This is a hack because I just cannot figure out how to do this
        natively in SQLAlchemy.

        :return: count of accounts with stale data
        :rtype: int
        """
        if sess is None:
            sess = db_session
        return sum(
            1 if a.is_stale else 0 for a in sess.query(
                Account).filter(Account.is_active.__eq__(True)).all()
        )

    @staticmethod
    def budget_account_sum(sess=None):
        """
        Return the sum of current balances for all is_budget_source accounts.

        :return: Combined balance of all budget source accounts
        :rtype: float
        """
        if sess is None:
            sess = db_session
        sum = Decimal('0.0')
        for acct in sess.query(Account).filter(
                Account.is_budget_source.__eq__(True),
                Account.is_active.__eq__(True)
        ):
            if acct.balance is not None:
                sum += acct.balance.ledger
        return sum

    @staticmethod
    def budget_account_unreconciled(sess=None):
        """
        Return the sum of unreconciled txns for all is_budget_source accounts.

        :return: Combined unreconciled amount of all budget source accounts
        :rtype: float
        """
        if sess is None:
            sess = db_session
        sum = Decimal('0.0')
        for acct in sess.query(Account).filter(
                Account.is_budget_source.__eq__(True),
                Account.is_active.__eq__(True)
        ):
            sum += acct.unreconciled_sum
        return sum

    @staticmethod
    def standing_budgets_sum(sess=None):
        """
        Return the sum of current balances of all standing budgets.

        :return: sum of current balances of all standing budgets
        :rtype: float
        """
        if sess is None:
            sess = db_session
        res = sess.query(func.sum(Budget.current_balance)).filter(
            Budget.is_periodic.__eq__(False),
            Budget.is_active.__eq__(True)
        ).all()[0][0]
        if res is None:
            return 0
        return res

    @staticmethod
    def pp_sum(sess=None):
        """
        Return the overall allocated sum for the current payperiod minus the
        sum of all reconciled Transactions for the pay period.

        :return: overall allocated sum for the current pay period minus the sum
          of all reconciled Transactions for the pay period.
        :rtype: float
        """
        if sess is None:
            sess = db_session
        pp = BiweeklyPayPeriod.period_for_date(dtnow(), sess)
        allocated = pp.overall_sums['allocated']
        spent = pp.overall_sums['spent']
        logger.debug('PayPeriod=%s; allocated=%s; spent=%s',
                     pp, allocated, spent)
        return allocated - spent

    @staticmethod
    def num_unreconciled_ofx(sess=None):
        """
        Return the number of unreconciled OFXTransactions.

        :return: number of unreconciled OFXTransactions
        :rtype: int
        """
        if sess is None:
            sess = db_session
        return OFXTransaction.unreconciled(sess).count()

    @staticmethod
    def get_notifications():
        """
        Return all notifications that should be displayed at the top of pages,
        as a list in the order they should appear. Each list item is a dict
        with keys "classes" and "content", where classes is the string that
        should appear in the notification div's "class" attribute, and content
        is the string content of the div.
        """
        res = []
        num_stale = NotificationsController.num_stale_accounts()
        if num_stale > 0:
            a = 'Accounts'
            if num_stale == 1:
                a = 'Account'
            res.append({
                'classes': 'alert alert-danger',
                'content': '%d %s with stale data. <a href="/accounts" '
                           'class="alert-link">View Accounts</a>.' % (num_stale,
                                                                      a)
            })
        accounts_bal = NotificationsController.budget_account_sum()
        unrec_amt = NotificationsController.budget_account_unreconciled()
        standing_bal = NotificationsController.standing_budgets_sum()
        curr_pp = NotificationsController.pp_sum()
        logger.info('accounts_bal=%s standing_bal=%s curr_pp=%s unrec=%s',
                    accounts_bal, standing_bal, curr_pp, unrec_amt)
        bal_sum = standing_bal + curr_pp + unrec_amt
        if accounts_bal < bal_sum:
            res.append({
                'classes': 'alert alert-danger',
                'content': 'Combined balance of all <a href="/accounts">'
                           'budget-funding accounts</a> '
                           '(%s) is less than all allocated funds total of '
                           '%s (%s <a href="/budgets">standing budgets</a>; '
                           '%s <a href="/pay_period_for">current pay '
                           'period remaining</a>; %s <a href="/reconcile">'
                           'unreconciled</a>)!'
                           '' % (
                               fmt_currency(accounts_bal),
                               fmt_currency(bal_sum),
                               fmt_currency(standing_bal),
                               fmt_currency(curr_pp),
                               fmt_currency(unrec_amt)
                           )
            })
        elif accounts_bal > bal_sum:
            res.append({
                'classes': 'alert alert-info',
                'content': 'Combined balance of all <a href="/accounts">'
                           'budget-funding accounts</a> '
                           '(%s) is more than all allocated funds total of '
                           '%s (%s <a href="/budgets">standing budgets</a>; '
                           '%s <a href="/pay_period_for">current pay '
                           'period remaining</a>; %s <a href="/reconcile">'
                           'unreconciled</a>)!'
                           '' % (
                               fmt_currency(accounts_bal),
                               fmt_currency(bal_sum),
                               fmt_currency(standing_bal),
                               fmt_currency(curr_pp),
                               fmt_currency(unrec_amt)
                           )
            })
        unreconciled_ofx = NotificationsController.num_unreconciled_ofx()
        if unreconciled_ofx > 0:
            res.append({
                'classes': 'alert alert-warning unreconciled-alert',
                'content': '%s <a href="/reconcile" class="alert-link">'
                           'Unreconciled OFXTransactions'
                           '</a>.' % unreconciled_ofx
            })
        return res



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from jinja2.runtime import Undefined
from humanize import naturaltime

from biweeklybudget.utils import dtnow, fmt_currency
from biweeklybudget.flaskapp.app import app
from biweeklybudget.models.account import AcctType


@app.template_filter('period_panel_color')
def period_panel_color_filter(x):
    """
    Given the remaining amount for a pay period, return "red" if less than
    zero, "yellow" if less than 100, or otherwise "green".

    :param x: PayPeriod remaining amount
    :type x: float
    :return: "red", "yellow" or "green"
    :rtype: str
    """
    if x <= 0:
        return "red"
    if x < 100:
        return "yellow"
    return "green"


@app.template_filter('dateymd')
def dateymd_filter(dt):
    """
    Format a datetime using %Y-%m-%d

    :param dt: datetime to format
    :type dt: datetime.datetime
    :return: formatted date
    :rtype: str
    """
    try:
        return dt.strftime('%Y-%m-%d')
    except AttributeError:
        return ''


@app.template_filter('isodate')
def isodate_filter(dt):
    """
    Format a datetime using %Y-%m-%d %H:%M:%S

    :param dt: datetime to format
    :type dt: datetime.datetime
    :return: formatted date
    :rtype: str
    """
    try:
        return dt.strftime('%Y-%m-%d %H:%M:%S')
    except AttributeError:
        return ''


@app.template_filter('ago')
def ago_filter(dt):
    """
    Format a datetime using humanize.naturaltime, "ago"

    :param dt: datetime to compare to now
    :type dt: datetime.datetime
    :return: ago string
    :rtype: str
    """
    if dt == '' or dt is None or isinstance(dt, Undefined):
        return ''
    return naturaltime(dtnow() - dt)


@app.template_filter('dollars')
def dollars_filter(x):
    """
    Format as currency using :py:func:`~.utils.fmt_currency`.

    :param x: currency amount, int, float, decimal, etc.
    :return: formatted currency
    :rtype: str
    """
    if x == '' or x is None or isinstance(x, Undefined):
        return ''
    return fmt_currency(x)


@app.template_filter('reddollars')
def reddollars_filter(x):
    """
    Return a string similar to ``dollars_filter`` but in red text
    if negative.

    :param x: dollar amount, int, float, decimal, etc.
    :return: formatted currency
    :rtype: str
    """
    s = dollars_filter(x)
    if x < 0:
        s = '<span class="text-danger">%s</span>' % s
    return s


@app.template_filter('pluralize')
def pluralize_filter(word, number=1):
    """
    If number is greater than one, return word with an "s" appended, else
    return word unmodified.

    :param word: the word to pluralize or not
    :type word: string
    :param number: the number to check for greater-than-one-ness
    :type number: int
    :return: word, pluralized or not
    :rtype: str
    """
    if number > 1:
        return word + 's'
    return word


@app.template_filter('acct_icon')
def acct_icon_filter(acct):
    """
    Given an Account, return the proper classes for an account type icon for it.

    :param acct: the account
    :type acct: biweeklybudget.models.account.Account
    :return: string icon classes
    :rtype: str
    """
    if acct.acct_type == AcctType.Bank:
        return 'fa fa-bank fa-fw'
    if acct.acct_type == AcctType.Credit:
        return 'fa fa-credit-card fa-fw'
    if acct.acct_type == AcctType.Investment:
        return 'glyphicon glyphicon-piggy-bank'
    if acct.acct_type == AcctType.Cash:
        return 'fa fa-dollar fa-fw'


@app.template_filter('decimal_to_percent')
def decimal_to_percent(d):
    if d is None:
        return ''
    d = d * 100
    return '%.2f%%' % d


@app.template_filter('dict_to_class_args')
def dict_to_class_args(j):
    if 'args' not in j and 'kwargs' not in j:
        return ''
    args = j.get('args', [])
    s = ', '.join(['%s' % a for a in args])
    kwargs = j.get('kwargs', {})
    s += ', '.join(['%s=%s' % (k, kwargs[k]) for k in kwargs.keys()])
    return s


@app.template_filter('monthsyears')
def monthsyears(num):
    if num < 12:
        return '%f months' % num
    return '%.1f years' % (num / 12.0)


@app.template_filter('budgetCell')
def budget_cell_filter(d):
    """
    Given a dictionary of budget IDs to names and amounts like that returned by
    :py:meth:`~.BiweeklyPayPeriod._dict_for_trans`, return the ``<td>`` content
    for those budgets.
    """
    if len(d) == 1:
        foo = list(d.keys())[0]
        return '<a href="/budgets/%d">%s</a>' % (foo, d[foo]['name'])
    items = [
        '<a href="/budgets/%d">%s</a> (%s)' % (
            k, d[k]['name'], dollars_filter(d[k]['amount'])
        ) for k in sorted(d.keys(), key=lambda x: d[x]['amount'], reverse=True)
    ]
    return '<br />'.join(items)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from datetime import date, datetime
from time import mktime
from json import JSONEncoder
from decimal import Decimal
import enum

from flask.json.provider import DefaultJSONProvider


class MagicJSONEncoder(JSONEncoder):
    """
    Customized JSONEncoder class that uses ``as_dict`` properties on objects
    to encode them.
    """

    def default(self, o):
        if hasattr(o, 'as_dict') and isinstance(type(o).as_dict, property):
            d = o.as_dict
            d['class'] = o.__class__.__name__
            return d
        if isinstance(o, datetime):
            return {
                'class': 'datetime',
                'str': o.strftime('%Y-%m-%d %H:%M:%S'),
                'ts': mktime(o.timetuple()),
                'year': o.year,
                'month': o.month,
                'date': o.day,
                'hour': o.hour,
                'minute': o.minute,
                'second': o.second,
                'tzname': o.strftime('%Z'),
                'tzoffset': o.strftime('%z'),
                'ymdstr': o.strftime('%Y-%m-%d')
            }
        if isinstance(o, date):
            return {
                'class': 'date',
                'str': o.strftime('%Y-%m-%d'),
                'ts': mktime(o.timetuple()),
                'year': o.year,
                'month': o.month,
                'date': o.day
            }
        if isinstance(o, Decimal):
            # Normalize decimal to remove trailing zeros, then convert to float
            # This ensures '100.0000' becomes 100, '10.23' stays 10.23
            return float(o.normalize())
        # Handle enum types
        if isinstance(o, enum.Enum):
            return o.value
        return super(MagicJSONEncoder, self).default(o)


class MagicJSONProvider(DefaultJSONProvider):
    """
    Flask 3.x JSON provider that uses MagicJSONEncoder.
    """

    def default(self, o):
        if hasattr(o, 'as_dict') and isinstance(type(o).as_dict, property):
            d = o.as_dict
            d['class'] = o.__class__.__name__
            return d
        if isinstance(o, datetime):
            return {
                'class': 'datetime',
                'str': o.strftime('%Y-%m-%d %H:%M:%S'),
                'ts': mktime(o.timetuple()),
                'year': o.year,
                'month': o.month,
                'date': o.day,
                'hour': o.hour,
                'minute': o.minute,
                'second': o.second,
                'tzname': o.strftime('%Z'),
                'tzoffset': o.strftime('%z'),
                'ymdstr': o.strftime('%Y-%m-%d')
            }
        if isinstance(o, date):
            return {
                'class': 'date',
                'str': o.strftime('%Y-%m-%d'),
                'ts': mktime(o.timetuple()),
                'year': o.year,
                'month': o.month,
                'date': o.day
            }
        if isinstance(o, Decimal):
            # Normalize decimal to remove trailing zeros, then convert to float
            # This ensures '100.0000' becomes 100, '10.23' stays 10.23
            return float(o.normalize())
        # Handle enum types
        if isinstance(o, enum.Enum):
            return o.value
        return super().default(o)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""
import logging

from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.txn_reconcile import TxnReconcile

logger = logging.getLogger(__name__)


def do_budget_transfer(db_sess, txn_date, amount, account,
                       from_budget, to_budget, notes=None):
    """
    Transfer a given amount from ``from_budget`` to ``to_budget`` on
    ``txn_date``. This method does NOT commit database changes. There are places
    where we rely on this function not committing changes.

    :param db_sess: active database session to use for queries
    :type db_sess: sqlalchemy.orm.session.Session
    :param txn_date: date to make the transfer Transactions on
    :type txn_date: datetime.date
    :param amount: amount of money to transfer
    :type amount: float
    :param account:
    :type account: biweeklybudget.models.account.Account
    :param from_budget:
    :type from_budget: biweeklybudget.models.budget_model.Budget
    :param to_budget:
    :type to_budget: biweeklybudget.models.budget_model.Budget
    :param notes: Notes to add to the Transaction
    :type notes: str
    :return: list of Transactions created for the transfer
    :rtype: :py:obj:`list` of :py:class:`~.Transaction` objects
    """
    desc = 'Budget Transfer - %s from %s (%d) to %s (%d)' % (
        amount, from_budget.name, from_budget.id, to_budget.name,
        to_budget.id
    )
    logger.info(desc)
    t1 = Transaction(
        date=txn_date,
        budget_amounts={from_budget: amount},
        budgeted_amount=amount,
        description=desc,
        account=account,
        notes=notes,
        planned_budget=from_budget
    )
    db_sess.add(t1)
    t2 = Transaction(
        date=txn_date,
        budget_amounts={to_budget: (-1 * amount)},
        budgeted_amount=(-1 * amount),
        description=desc,
        account=account,
        notes=notes,
        planned_budget=to_budget
    )
    db_sess.add(t2)
    t1.transfer = t2
    db_sess.add(t1)
    t2.transfer = t1
    db_sess.add(t2)
    db_sess.add(TxnReconcile(
        transaction=t1,
        note=desc
    ))
    db_sess.add(TxnReconcile(
        transaction=t2,
        note=desc
    ))
    return [t1, t2]



from __future__ import absolute_import
from __future__ import unicode_literals
try:
    # python 3
    from http.client import HTTPSConnection
except ImportError:
    # python 2
    from httplib import HTTPSConnection
import logging
import time
try:
    # python 3
    from urllib.parse import splittype, splithost
except ImportError:
    # python 2
    from urllib import splittype, splithost
import uuid

DEFAULT_APP_ID = 'QWIN'
DEFAULT_APP_VERSION = '2500'
DEFAULT_OFX_VERSION = '102'
DEFAULT_USER_AGENT = 'httpclient'
DEFAULT_ACCEPT = '*/*, application/x-ofx'

LINE_ENDING = "\r\n"


def ofx_uid():
    return str(uuid.uuid4().hex)


class Client:
    """This communicates with the banks via the OFX protocol

    :param institution: institution to connect to
    :type institution: :py:class:`ofxclient.Institution`
    :param id: client id (optional need for OFX version >= 103)
    :type id: string
    :param app_id: OFX app id
    :type app_id: string
    :param app_version: OFX app version
    :type app_version: string
    :param ofx_version: OFX spec version
    :type ofx_version: string
    :param user_agent: Value to send for User-Agent HTTP header. Leave as
      None to send default. Set to False to not send User-Agent header.
    :type user_agent: str, None or False
    :param accept: Value to send for Accept HTTP header. Leave as
      None to send default. Set to False to not send User-Agent header.
    :type accept: str, None or False
    """

    def __init__(
        self,
        institution,
        id=ofx_uid(),
        app_id=DEFAULT_APP_ID,
        app_version=DEFAULT_APP_VERSION,
        ofx_version=DEFAULT_OFX_VERSION,
        user_agent=DEFAULT_USER_AGENT,
        accept=DEFAULT_ACCEPT
    ):
        self.institution = institution
        self.id = id
        self.app_id = app_id
        self.app_version = app_version
        self.ofx_version = ofx_version
        self.user_agent = user_agent
        self.accept = accept
        # used when serializing Institutions
        self._init_args = {
            'id': self.id,
            'app_id': self.app_id,
            'app_version': self.app_version,
            'ofx_version': self.ofx_version,
            'user_agent': self.user_agent,
            'accept': self.accept
        }
        self.cookie = 3

    @property
    def init_args(self):
        """
        Return a dict of the arguments used to initialize this client,
        suitable for use when serializing an Institution.

        :return: constructor arguments
        :rtype: dict
        """
        return self._init_args

    def authenticated_query(
        self,
        with_message=None,
        username=None,
        password=None
    ):
        """Authenticated query

        If you pass a 'with_messages' array those queries will be passed along
        otherwise this will just be an authentication probe query only.
        """
        u = username or self.institution.username
        p = password or self.institution.password

        contents = ['OFX', self._signOn(username=u, password=p)]
        if with_message:
            contents.append(with_message)
        return LINE_ENDING.join([self.header(), _tag(*contents)])

    def bank_account_query(self, number, date, account_type, bank_id):
        """Bank account statement request"""
        return self.authenticated_query(
            self._bareq(number, date, account_type, bank_id)
        )

    def credit_card_account_query(self, number, date):
        """CC Statement request"""
        return self.authenticated_query(self._ccreq(number, date))

    def brokerage_account_query(self, number, date, broker_id):
        return self.authenticated_query(
            self._invstreq(broker_id, number, date))

    def account_list_query(self, date='19700101000000'):
        return self.authenticated_query(self._acctreq(date))

    def post(self, query):
        """
        Wrapper around ``_do_post()`` to handle accounts that require
        sending back session cookies (``self.set_cookies`` True).
        """
        res, response = self._do_post(query)
        cookies = res.getheader('Set-Cookie', None)
        if len(response) == 0 and cookies is not None and res.status == 200:
            logging.debug('Got 0-length 200 response with Set-Cookies header; '
                          'retrying request with cookies')
            _, response = self._do_post(query, [('Cookie', cookies)])
        return response

    def _do_post(self, query, extra_headers=[]):
        """
        Do a POST to the Institution.

        :param query: Body content to POST (OFX Query)
        :type query: str
        :param extra_headers: Extra headers to send with the request, as a list
          of (Name, Value) header 2-tuples.
        :type extra_headers: list
        :return: 2-tuple of (HTTPResponse, str response body)
        :rtype: tuple
        """
        i = self.institution
        logging.debug('posting data to %s' % i.url)
        garbage, path = splittype(i.url)
        host, selector = splithost(path)
        h = HTTPSConnection(host, timeout=60)
        # Discover requires a particular ordering of headers, so send the
        # request step by step.
        h.putrequest('POST', selector, skip_host=True,
                     skip_accept_encoding=True)
        headers = [
            ('Content-Type', 'application/x-ofx'),
            ('Host', host),
            ('Content-Length', len(query)),
            ('Connection', 'Keep-Alive')
        ]
        if self.accept:
            headers.append(('Accept', self.accept))
        if self.user_agent:
            headers.append(('User-Agent', self.user_agent))
        for ehname, ehval in extra_headers:
            headers.append((ehname, ehval))
        logging.debug('---- request headers ----')
        for hname, hval in headers:
            logging.debug('%s: %s', hname, hval)
            h.putheader(hname, hval)
        logging.debug('---- request body (query) ----')
        logging.debug(query)
        h.endheaders(query.encode())
        res = h.getresponse()
        response = res.read().decode('ascii', 'ignore')
        logging.debug('---- response ----')
        logging.debug(res.__dict__)
        logging.debug('Headers: %s', res.getheaders())
        logging.debug(response)
        res.close()
        return res, response

    def next_cookie(self):
        self.cookie += 1
        return str(self.cookie)

    def header(self):
        parts = [
            "OFXHEADER:100",
            "DATA:OFXSGML",
            "VERSION:%d" % int(self.ofx_version),
            "SECURITY:NONE",
            "ENCODING:USASCII",
            "CHARSET:1252",
            "COMPRESSION:NONE",
            "OLDFILEUID:NONE",
            "NEWFILEUID:"+ofx_uid(),
            ""
        ]
        return LINE_ENDING.join(parts)

    """Generate signon message"""
    def _signOn(self, username=None, password=None):
        i = self.institution
        u = username or i.username
        p = password or i.password
        fidata = [_field("ORG", i.org)]
        if i.id:
            fidata.append(_field("FID", i.id))

        client_uid = ''
        if str(self.ofx_version) == '103':
            client_uid = _field('CLIENTUID', self.id)

        return _tag("SIGNONMSGSRQV1",
                    _tag("SONRQ",
                         _field("DTCLIENT", now()),
                         _field("USERID", u),
                         _field("USERPASS", p),
                         _field("LANGUAGE", "ENG"),
                         _tag("FI", *fidata),
                         _field("APPID", self.app_id),
                         _field("APPVER", self.app_version),
                         client_uid
                         ))

    def _acctreq(self, dtstart):
        req = _tag("ACCTINFORQ", _field("DTACCTUP", dtstart))
        return self._message("SIGNUP", "ACCTINFO", req)

# this is from _ccreq below and reading page 176 of the latest OFX doc.
    def _bareq(self, acctid, dtstart, accttype, bankid):
        req = _tag("STMTRQ",
                   _tag("BANKACCTFROM",
                        _field("BANKID", bankid),
                        _field("ACCTID", acctid),
                        _field("ACCTTYPE", accttype)),
                   _tag("INCTRAN",
                        _field("DTSTART", dtstart),
                        _field("INCLUDE", "Y")))
        return self._message("BANK", "STMT", req)

    def _ccreq(self, acctid, dtstart):
        req = _tag("CCSTMTRQ",
                   _tag("CCACCTFROM", _field("ACCTID", acctid)),
                   _tag("INCTRAN",
                        _field("DTSTART", dtstart),
                        _field("INCLUDE", "Y")))
        return self._message("CREDITCARD", "CCSTMT", req)

    def _invstreq(self, brokerid, acctid, dtstart):
        req = _tag("INVSTMTRQ",
                   _tag("INVACCTFROM",
                        _field("BROKERID", brokerid),
                        _field("ACCTID", acctid)),
                   _tag("INCTRAN",
                        _field("DTSTART", dtstart),
                        _field("INCLUDE", "Y")),
                   _field("INCOO", "Y"),
                   _tag("INCPOS",
                        _field("DTASOF", now()),
                        _field("INCLUDE", "Y")),
                   _field("INCBAL", "Y"))
        return self._message("INVSTMT", "INVSTMT", req)

    def _message(self, msgType, trnType, request):
        return _tag(msgType+"MSGSRQV1",
                    _tag(trnType+"TRNRQ",
                         _field("TRNUID", ofx_uid()),
                         _field("CLTCOOKIE", self.next_cookie()),
                         request))


def _field(tag, value):
    return "<"+tag+">"+value


def _tag(tag, *contents):
    return LINE_ENDING.join(['<'+tag+'>']+list(contents)+['</'+tag+'>'])


def now():
    return time.strftime("%Y%m%d%H%M%S", time.localtime())
